# MMAPPR version 0.83. Copyright 2012-2013, University of Utah.
#===============================================================================
# MMAPPR: Mutation Mapping Analysis Pipeline for Pooled RNA-seq
# Jonathon Hill, Bradley Demarest, Brent Bisgrove, Bushra Gorsi, H. Joseph Yost
# Department of Neurobiology and Anatomy, University of Utah
#===============================================================================

require(parallel)
args <- commandArgs(trailingOnly=T)

power <- as.numeric(args[3])
m=args[4]
md <- as.numeric(args[5]) - 1
c1 <- makeCluster(as.numeric(args[6]))
chrname <- args[7]

loessopt <- function (s, e, p, m) {
	# extract values from loess object
	x <- try(loess(e ~ p, span=s, degree=1, family="symmetric", surface='direct'), silent=T)
	if(class(x)=="try-error"){return(NA)}
	span <- x$pars$span
	n <- x$n
	traceL <- x$trace.hat
	sigma2 <- sum( x$residuals^2 ) / (n-1)
	delta1 <- x$one.delta
	delta2 <- x$two.delta
	enp <- x$enp
	if(m=="aicc"){return(log(sigma2) + 1 + 2* (2*(traceL+1)) / (n-traceL-2))}
	if(m=="gcv"){return(n*sigma2 / (n-traceL)^2)}
}

cat("Importing files.\n") 
cols<-c('chr','pos', 'ref', 'cov','A','C','G','T')
coltypes <- c('character',  'integer', 'character', 'integer', 'numeric', 'numeric', 'numeric', 'numeric')

con <- read.table(args[1], colClasses=coltypes, comment.char='', sep="\t", header=T) #need to set control file


con$A<-con$A/con$cov
con$C<-con$C/con$cov
con$G<-con$G/con$cov
con$T<-con$T/con$cov

mut <- read.table(args[2], colClasses=coltypes, comment.char='', sep="\t", header=T) #need to set mut file


mut$A<-mut$A/mut$cov
mut$C<-mut$C/mut$cov
mut$G<-mut$G/mut$cov
mut$T<-mut$T/mut$cov

comb <- merge(con[con$A<.95 & con$G<.95 & con$C<.95 & con$T<.95, ], mut, by=c("chr","pos"))
comb$euc<-sqrt((comb$A.y-comb$A.x)^2+(comb$C.y-comb$C.x)^2+(comb$G.y-comb$G.x)^2+(comb$T.y-comb$T.x)^2)
comb2 <- comb
comb <- comb[comb$cov.x > md & comb$cov.y > md,]


comb$nums <- as.numeric(gsub("\\D", "", comb$chr))
comb <- comb[order(comb$nums, comb$chr, comb$pos), ]
chrs <- unique(comb$chr)
chrs <- chrs[grep(chrname, chrs)]

cat("Processing each chromosome:\nChr\tSnps\tSpan\n")

for(chr in chrs) {
	e = comb$euc[comb$chr == chr]^power
	p = comb$pos[comb$chr == chr]
	spanresults.span <- NULL
	spanresults.aicc <- NULL
	if(as.integer(length(p)) < 50){
		cat(chr, " has less than 50 snps (n=", length(p), ")\n", sep="")
		next
	}
	spanresults.span <- seq(round(50/length(p),digits=3), 1, .001)
	spanresults.aicc <- parSapply(c1, spanresults.span, loessopt, e=e, p=p, m=m)
	usespan <- spanresults.span[spanresults.aicc==min(spanresults.aicc, na.rm=T)]
	lo <- loess(e ~ p, span = usespan[1], degree=1, family='symmetric', surface='direct') #designate usespan index incase of tie.
	comb$fitted[comb$chr==chr] <- lo$fitted
	comb$unfitted[comb$chr==chr] <- lo$y
	cat(chr, length(p), round(usespan[1], digits=3), "\n", sep="\t")
}
stopCluster(c1)
comb <- comb[!is.na(comb$fitted),]

breaks <- NULL
for (row in 1:(nrow(comb)-1)){
	if (comb$chr[row] != comb$chr[row+1]) {
		breaks <- append(breaks, row+1)
	}
}
breaks <- append(breaks, nrow(comb))

lastbreak <- 1
labelpos <- NULL
plot.df <- data.frame(cbind(NA, NA, NA, NA, NA))
colnames(plot.df) <- c('chr','pos','euc','fitted','unfitted')
for( b in breaks){
	plot.df <- rbind(plot.df, comb[lastbreak:b,c('chr','pos','euc','fitted','unfitted')],c(NA, NA, NA, NA, NA))
	labelpos <- append(labelpos, ((b+1)-lastbreak)/2+lastbreak)
	lastbreak <- b
}

pdf("snpmapresults/Rplots.pdf", width=11, height=8.5)
par(mfrow=c(2,1))
plot(1:nrow(plot.df), plot.df$fitted, type='l', ylim=c(min(plot.df$fitted, na.rm=T),1.1*max(plot.df$fitted, na.rm=T)), ylab=substitute("ED"^p~ ~"(Loess fit)", list(p=power)), xaxt='n', xaxs='i', xlab="Chromosome" )
abline(v=(breaks[1:length(breaks)-1]+2), col="grey")
mtext(unique(gsub("\\D", "", plot.df$chr[!is.na(plot.df$chr)])), at = labelpos, side=1, cex=.6)
plot(1:nrow(plot.df), plot.df$unfitted, pch=16, cex=.8, col="#999999AA", ylim=c(min(plot.df$unfitted, na.rm=T),1.1*max(plot.df$unfitted, na.rm=T)), ylab=substitute("ED"^p, list(p=power)), xaxt='n', xaxs='i', xlab="Chromosome" )
abline(v=(breaks[1:length(breaks)-1]+2), col="grey")
mtext(unique(gsub("\\D", "", plot.df$chr[!is.na(plot.df$chr)])), at = labelpos, side=1, cex=.6)

cutoff <- 3*(sd(comb$fitted)+median(comb$fitted))
cat("Using", round(cutoff, digits=4), "as cutoff.\n")
starts <- NULL
stops <- NULL
peakchr <- NULL
for(chr in chrs){
	chr.active <- plot.df[plot.df$chr==chr & !is.na(plot.df$fitted),]
	for(j in 1:nrow(chr.active)){
		if (j==0){next}
		if (is.na(chr.active$fitted[j])){next}
		if (j==nrow(chr.active) & chr.active$fitted[j] > cutoff) {stops <- append(stops, chr.active$pos[j]); next}
		if (j==1 & chr.active$fitted[j] > cutoff) {peakchr <- append(peakchr, chr); starts <- append(starts, 0); next}
		if (j!=nrow(chr.active) & chr.active$fitted[j] < cutoff & chr.active$fitted[j+1] > cutoff) {peakchr <- append(peakchr, chr); starts <- append(starts, chr.active$pos[j+1]); next}
		if (j!=nrow(chr.active) & chr.active$fitted[j] > cutoff & chr.active$fitted[j+1] < cutoff) {stops <- append(stops, chr.active$pos[j]); next}		
	}
}	

peak.df <- data.frame(cbind(peakchr,starts,stops), stringsAsFactors = F)
colnames(peak.df) <- c('chr','starts','stops')
peakmax <- NULL
if (nrow(peak.df) < 1) {
	cat("No peaks found")
	dev.off()
	quit()
}
for(i in 1:nrow(peak.df)){
	peak.df$peakmax[i] <- comb$pos[comb$fitted==max(comb$fitted[comb$pos > peak.df$starts[i] & comb$pos < peak.df$stops[i]], na.rm=T) & comb$pos > peak.df$starts[i] & comb$pos < peak.df$stops[i]]
	peak.df$height[i] <- max(comb$fitted[comb$pos > peak.df$starts[i] & comb$pos < peak.df$stops[i]], na.rm=T) - median(comb$fitted)
}
cat("Peaks:\n")
write.table(peak.df, quote=FALSE, row.names=FALSE, sep="\t")
for(chr in unique(peak.df$chr)){
	plot(plot.df$pos[plot.df$chr==chr]/1000000, plot.df$fitted[plot.df$chr==chr], type='l', ylim=c(min(plot.df$fitted, na.rm=T),1.1*max(plot.df$fitted, na.rm=T)), ylab=substitute("ED"^p~ ~"(Loess fit)", list(p=power)), xlab=paste(chr,"Base Position (MB)"), xaxs='i' )
	chr.a <- transform(peak.df[peak.df$chr==chr,], starts = as.integer(starts), stops = as.integer(stops))
	for(i in 1:nrow(chr.a)){
		shade.x <- c(chr.a$starts[i],plot.df$pos[plot.df$pos >= chr.a$starts[i] & plot.df$pos <= chr.a$stops[i] & plot.df$chr==chr & !is.na(plot.df$fitte)],chr.a$stops[i])
		shade.y <- c(-5, plot.df$fitted[plot.df$pos >= chr.a$starts[i] & plot.df$pos <= chr.a$stops[i] & plot.df$chr==chr & !is.na(plot.df$fitted)], -5)
		polygon(shade.x/1000000,shade.y, col=rgb(0,1,1,.75), border=NA)
	}
	plot(plot.df$pos[plot.df$chr==chr]/1000000, plot.df$unfitted[plot.df$chr==chr], pch=16, cex=.6, ylim=c(min(plot.df$unfitted, na.rm=T),1.1*max(plot.df$unfitted, na.rm=T)), ylab=substitute("ED"^p, list(p=power)), xlab=paste(chr,"Base Position (MB)"), xaxs='i' )
}
noreport <- dev.off()

cat("Making allele file.", "\n")

genotype <- function(a, c, g, t){
	maxbase <- max(c( a, c, g, t))
	if(maxbase == a){return("A")}
	if(maxbase == c){return("C")}
	if(maxbase == g){return("G")}
	if(maxbase == t){return("T")}
}

alleler <- data.frame(t(rep(NA, 5)))
names(alleler) <- c("chr", "pos", "pos.1", "call", "notes")

for (i in 1:nrow(peak.df)) {
	tempalleler <- comb2[comb2$chr==peak.df$chr[i] & comb2$pos > peak.df$starts[i] & comb2$pos < peak.df$stops[i] & (comb2$A.y>.75 | comb2$G.y>.75 | comb2$C.y>.75 | comb2$T.y>.75) & comb2$cov.y > 1 & comb2$euc > .5, c('chr','pos','pos', 'ref.y', 'cov.y', 'A.y', 'C.y', 'G.y', 'T.y')]
	names(tempalleler) <- c("chr", "pos", "pos.1", "ref", "cov", "A", "C", "G", "T")
	tempalleler$pos <- tempalleler$pos-1
	
	tempalleler$call <- mapply(genotype, tempalleler$A, tempalleler$C, tempalleler$G, tempalleler$T)
	tempalleler <- tempalleler[tempalleler$call != tempalleler$ref, ]
	
	tempalleler$notes <- mapply(function(x,a,c,g,t) paste("Coverage:A:C:G:T = ", x, ":", a, ":", c, ":", g, ":", t, sep=""), tempalleler$cov, as.integer(tempalleler$A*tempalleler$cov), as.integer(tempalleler$C*tempalleler$cov), as.integer(tempalleler$G*tempalleler$cov), as.integer(tempalleler$T*tempalleler$cov))
	alleler <- rbind(alleler, tempalleler[c('chr', 'pos', 'pos.1', 'call','notes')])
}

alleler <- alleler[!is.na(alleler$chr), ]

write.table(alleler, file="snpmapresults/putativesnps.txt", sep = "\t", row.names=F, col.names=F,quote=F)

gc()
